from sympy import (Symbol, symbols, Matrix, sin, cos, asin, diff, sqrt, S,
                   diag, Eq, hessian, Function, flatten, Tuple, im, pi, latex,
                   dsolve, solve, fraction, factorial, Subs, Number, oo, Abs,
                   N, solveset, atan)

from sympy.physics.mechanics import dynamicsymbols, ReferenceFrame, Point
from sympy.physics.vector import vpprint, vlatex
from ...dynamics import LagrangesDynamicSystem, HarmonicOscillator, mech_comp

from ..elements import MaterialPoint, Spring, GravitationalForce, Disk, RigidBody2D, Damper, PID, Excitation, Force, base_frame, base_origin, EngineMount
from ..continuous import ContinuousSystem, PlaneStressProblem
from dynpy.models.mechanics.tmd import TunedMassDamper

import base64
import random
import IPython as IP
import numpy as np
import inspect

from .principles import ComposedSystem, NonlinearComposedSystem, base_frame, base_origin, REPORT_COMPONENTS_LIST

import matplotlib.pyplot as plt

import pint
ureg = pint.UnitRegistry()

class YokeMechanism(ComposedSystem):

    scheme_name = 'crank_mechanismlow.jpg'
    real_name = 'crank_slider_real.jpg'

    
    _default_folder_path = "./dynpy/models/images/"
    
    I=Symbol('I', positive=True)
    r=Symbol('r', positive=True)
    h=Symbol('h', positive=True)
    a=Symbol('a', positive=True)
    phi=dynamicsymbols('\\varphi')
    omega=Symbol('omega',positive=True)
    
    def __init__(self,
                 I=None,
                 r=None,
                 h=None,
                 a=None,
                 phi=None,
                 omega=None,
                 ivar=None,
                 **kwargs):

        if I is not None: self.I = I
        if h is not None: self.h = h
        if r is not None: self.r = r
        if a is not None: self.a = a
        if phi is not None: self.phi = phi
        if omega is not None: self.omega = omega
        if ivar is not None: self.ivar = ivar
        self.qs = [self.phi]

        self._init_from_components(**kwargs)

    @property
    def components(self):
        components = {}
        I,h,r,a,phi,omega = self.I,self.h,self.r,self.a,self.phi,self.omega

        self._crank = MaterialPoint(I, pos1=phi, qs=[phi])

        components['crank'] = self._crank

        return components

    def _crank_horizontal_comp(self,subs=False):

        phi = self.phi
        r = self.r

        pAx = 0
        pBx = pAx + r*sin(phi)

        if subs is False:
            return pBx
        else:
            print(self._given_data)
            return float(pBx.subs(self._given_data))

    def _crank_vertical_comp(self,subs=False):

        phi = self.phi
        r = self.r
        h = self.h

        pAy = h
        pBy = pAy + r*cos(phi)

        if subs is False:
            return pBy
        else:
            return float(pBy.subs(self._given_data))

    def _guide_sin_comp(self,subs=False):

        pOx = 0
        pOy = 0
        pBx = self._crank_horizontal_comp(subs)
        pBy = self._crank_vertical_comp(subs)

        lBO = ((pBx-pOx)**2 + (pBy-pOy)**2)**0.5
        guide_angle = pBy/lBO

        return guide_angle

    def _guide_cos_comp(self,subs=False):

        pOx = 0
        pOy = 0
        pBx = self._crank_horizontal_comp(subs)
        pBy = self._crank_vertical_comp(subs)

        lBO = ((pBx-pOx)**2 + (pBy-pOy)**2)**0.5
        guide_angle = pBx/lBO

        return guide_angle

    def preview(self, example=False):


        if self._path:
            path = self._path

        else:

            plt.figure(figsize=(10,10))
#             plt.figure()


            #plt.xlim(-0.5,1)
            #plt.ylim(-0.25,1.25)
            plt.axis('equal')
            plt.grid(True)
            
            if (self._given_data)=={}:
            
                path = self.__class__._default_folder_path + self.__class__.scheme_name
            else:

                h_num=float(self.h.subs(self._given_data))

                pOx = 0
                pOy = 0
                plt.text(-0.05+(pOx),(-0.05+pOy),'O')

                pAx=0
                pAy=h_num
                plt.text(-0.05+(pAx),(+0.05+pAy),'A')

                plt.plot([pOx,pAx],[pOy,pAy],'b',linewidth=2)
                plt.text(-0.05+(pOx+pAx)/2,(pOy+pAy)/2,'h')

                pBx = self._crank_horizontal_comp(True)
                pBy = self._crank_vertical_comp(True)
                plt.text(+0.05+(pBx),(+0.00+pBy),'B')

                plt.plot([pBx,pAx],[pBy,pAy],'r',linewidth=2)
                plt.text((pBx+pAx)/2,0.05+(pBy+pAy)/2,'r')

                plt.plot([pOx,pBx],[pOy,pBy],'g',linewidth=2)
                plt.text((pOx+pBx)/2,0.05+(pOy+pBy)/2,'a')

                lBO = ((pBx-pOx)**2 + (pBy-pOy)**2)**0.5
                pDx = lBO*self._guide_cos_comp(True)*1.2
                pDy = lBO*self._guide_sin_comp(True)*1.2
                plt.plot([pOx,pDx],[pOy,pDy],'g',linewidth=2)
                plt.text(+0.05+(pDx),(+0.00+pDy),'D')

                path = self.__class__._default_folder_path + 'previews/' + self.__class__.__name__ + str(
                                            next(self.__class__._case_no)) + '.png'

                plt.savefig(path)
                self._path=path
                plt.show()
                plt.close()


#        print('check' * 100)
        print(self._path)
#        print('check' * 100)
        plt.close()

        with open(f"{path}", "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read())
        image_file.close()

        return IP.display.Image(base64.b64decode(encoded_string))

    def get_default_data(self):

        default_data_dict = {
            self.h: [0.6,0.6,0.55,0.5,0.45,0.4,0.5,0.55,0.6],
            self.r: [0.35,0.3,0.4,0.45,0.5,0.35,0.4,0.3,0.2],
            self.phi: [30*pi/180,35*pi/180,40*pi/180,45*pi/180,50*pi/180,55*pi/180,180*pi/180,275*pi/180,300*pi/180,35*pi/180]
        }
        return default_data_dict
    
    
    
class CrankMechanism(ComposedSystem):

    scheme_name = 'crank_mechanismlow.jpg'
    real_name = 'crank_slider_real.jpg'

    
    _default_folder_path = "./dynpy/models/images/"

    
    def __init__(self,
                 I=Symbol('I', positive=True),
                 a=Symbol('a', positive=True),
                 b=Symbol('b', positive=True),
                 alpha=dynamicsymbols('alpha'),
                 beta=dynamicsymbols('beta'),
                 **kwargs):

        self.I = I
        self.a = a
        self.b = b
        self.alpha = alpha
        self.beta = beta

        self.crank = MaterialPoint(I, alpha, qs=[alpha])
        composed_system = (self.crank)

        super().__init__(composed_system,**kwargs)


    def _crank_horizontal_comp(self,subs=False):

        pCx = a*cos(alpha)

        if subs is False:
            return pCx
        else:
            return float(pCx.subs(self._given_data))

    def _crank_vertical_comp(self,subs=False):

        pCy = a*sin(alpha)

        if subs is False:
            return pCy
        else:
            return float(pCy.subs(self._given_data))

    def _piston_horizontal_comp(self,subs=False):

        pDx = self._crank_horizontal_comp() + (b**2 - self._crank_vertical_comp())**0.5
        
        if subs is False:
            return pDx
        else:
            return float(pDx.subs(self._given_data))

    def _pistion_vertical_comp(self):

        pDy = 0
        return pDy
    

class CrankSystem(ComposedSystem):

    scheme_name = 'crank_mechanismlow.jpg'
    real_name = 'crank_slider_real.jpg'

    
    _default_folder_path = "./dynpy/models/images/"



    def _detail_scheme(self):

        self.preview()

        return self._path
    
    
    def preview(self, example=False):
        

        if self._path:
            path = self._path
            
        else:

            plt.figure(figsize=(10,10))


            plt.xlim(-0.5,1)
            plt.ylim(-0.25,1.25)
            plt.grid(True)
            
            if (self._given_data)=={}:
            
                path = self.__class__._default_folder_path + self.__class__.scheme_name
            else:
                

                
                h_num=float(self.h.subs(self._given_data))
                r_num=float(self.r.subs(self._given_data))
                phi_num=float(self.phi.subs(self._given_data))
                a_num=float(self.a.subs(self._given_data))
                b_num=float(self.b.subs(self._given_data))                
                
                pOx = 0
                pOy = 0
                plt.text(-0.05+(pOx),(-0.05+pOy),'O')
                
                pAx=0
                pAy=h_num
                plt.text(-0.05+(pAx),(+0.05+pAy),'A')
                
                plt.plot([pOx,pAx],[pOy,pAy],'b',linewidth=2)
                plt.text(-0.05+(pOx+pAx)/2,(pOy+pAy)/2,'h')
                
                pBx = pAx + r_num*np.sin(phi_num)
                pBy = pAy + r_num*np.cos(phi_num)
                plt.text(+0.05+(pBx),(+0.00+pBy),'B')
                
                plt.plot([pBx,pAx],[pBy,pAy],'r',linewidth=2)
                plt.text((pBx+pAx)/2,0.05+(pBy+pAy)/2,'r')
                
                lBO = ((pBx-pOx)**2 + (pBy-pOy)**2)**0.5
                
                p1x = 1.1*(h_num+r_num)*pBx/lBO
                p1y = 1.1*(h_num+r_num)*pBy/lBO
                
                plt.plot([p1x,pOx],[p1y,pOy],'g',linewidth=2)

                pCx = a_num*pBx/lBO
                pCy = a_num*pBy/lBO
                plt.text(+0.05+(pCx),(+0.05+pCy),'C')
                
                plt.plot([pOx,pCx],[pOy,pCy],'m',linewidth=2)
                plt.text(0.05+(pOx+pCx)/2,(pOy+pCy)/2,'a')
                
#                 pDx = pCx+(b_num**2 - pCy**2)**0.5        # poprawna wartość położenia D
                pDx = (pCx+(b_num**2 - pCy**2)**0.5)*1.15   # celowo wprowadzony błąd
                pDy = 0
                plt.text(+0.05+(pDx),(+0.00+pDy),'D')
                
                plt.plot([pCx,pDx],[pCy,pDy],'k',linewidth=2)
                plt.text((pCx+pDx)/2,0.05+(pCy+pDy)/2,'b')
            
                path = self.__class__._default_folder_path + 'previews/' + self.__class__.__name__ + str(
                                            next(self.__class__._case_no)) + '.png'

                plt.savefig(path)
                self._path=path

                plt.close()

        

#        print('check' * 100)
        print(self._path)
#        print('check' * 100)
        plt.close()

        with open(f"{path}", "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read())
        image_file.close()

        return IP.display.Image(base64.b64decode(encoded_string))
    
    
    
    def __init__(self,
                 I=Symbol('I', positive=True),
                 r=Symbol('r', positive=True),
                 h=Symbol('h', positive=True),
                 a=Symbol('a', positive=True),
                 b=Symbol('b', positive=True),
                 phi=dynamicsymbols('\\varphi'),
                 beta=dynamicsymbols('beta'),
#                  alpha=dynamicsymbols('alpha'),
                 **kwargs):

        self.I = I
        self.h=h
        self.r=r
        self.phi = phi
        self.beta = beta
#         self.alpha = alp
        self.a = a
        self.b = b

        self.crank = MaterialPoint(I, phi, qs=[phi])
        composed_system = (self.crank)

        super().__init__(composed_system,**kwargs)

        
    def _crank_horizontal_pos(self):
        return None
        
    @property
    def _dbeta(self):
        beta=atan(self.r/self.l*self.phi) #it's probably wrong - has to be checked
        return beta.diff(self.ivar)
    
    @property
    def _displacement_d(self):

        return -(-self.b*sqrt((self.h**2 - self.h**2*self.a**2/self.b**2 + 2*self.h*self.r*cos(self.phi) - 2*self.h*self.r*self.a**2*cos(self.phi)/self.b**2 + self.r**2 - self.r**2*self.a**2*cos(self.phi)**2/self.b**2)/(self.h**2 + 2*self.h*self.r*cos(self.phi) + self.r**2)) - self.r*self.a*sin(self.phi)/sqrt(self.h**2 + 2*self.h*self.r*cos(self.phi) + self.r**2))
    
    @property
    def _velocity_b2(self):
        return (self.phi.diff(self.ivar)*self.r)
    @property
    def _velocity_b2b3(self):
        return (sqrt(self.h**2 + 2*self.h*self.r*cos(self.phi) + self.r**2)).diff(self.ivar)
    @property
    def _velocity_b3(self):
        gamma=asin(self._velocity_b2b3/self._velocity_b2)
        return self._velocity_b2*cos(gamma)
    @property
    def _velocity_d(self):
        return self._displacement_d.diff(self.ivar)
    @property
    def _velocity_c(self):
        omega3=self._velocity_b3/sqrt(self.r**2 + self.h**2 - 2*self.r*self.h*cos(pi-self.phi))
        return omega3*self.a
    @property
    def _acceleration_b2(self):
        return self.phi.diff(self.ivar)**2*self.r
    @property
    def _acceleration_b3n(self):
        return (self._velocity_b3**2/sqrt(self.r**2 + self.h**2 - 2*self.r*self.h*cos(pi-self.phi)))
    @property
    def _acceleration_cn(self):
        return (self._velocity_b3**2*(self.a/(sqrt(self.r**2 + self.h**2 - 2*self.r*self.h*cos(pi-self.phi)))**2))
    @property
    def _acceleration_d(self):
        return self._velocity_d.diff(self.ivar)

    
    @property
    def _omega_3(self):
        return (self._velocity_b3/sqrt(self.r**2 + self.h**2 - 2*self.r*self.h*cos(pi-self.phi)))
    @property
    def linkage_ang_velocity(self):
        return self._dbeta

    def symbols_description(self):
        self.sym_desc_dict = {
            self.I: r'crank moment of inertia',
            # self.k_beam: r'Beam stiffness',
            # self.g: r'gravitational field acceleration'
        }

        return self.sym_desc_dict

    def get_default_data(self):
        E, I, l, m0, k0 = symbols('E I0 l_beam m_0 k_0', positive=True)
        
        default_data_dict = {
            self.I: [20 * I, 30 * I,60 * I,50 * I,40 * I,],
            self.h:[1,2,3],
            self.r:[0.1,0.2,0.3],
            self.a:[10],
            self.b:[20],
            self.phi : [10,20,30],
        }

        return default_data_dict


class SDOFWinchSystem(ComposedSystem):

    scheme_name = 'winch_complete.png'
    real_name = 'winch_cross_sec.png'

    I_s=Symbol('I_s', positive=True)
    I_k=Symbol('I_k', positive=True)
    I_1=Symbol('I_1', positive=True)
    i_1=Symbol('i_1', positive=True)
    I_2=Symbol('I_2', positive=True)
    i_2=Symbol('i_2', positive=True)
    I_3=Symbol('I_3', positive=True)
    I_4=Symbol('I_4', positive=True)
    I_b=Symbol('I_b', positive=True)
    A=Symbol('A', positive=True)
    B=Symbol('B', positive=True)
    M_r=Symbol('M_r')
    M_b=Symbol('M_b')
    I_r=Symbol('I_r')
    ivar=Symbol('t')
    phi=dynamicsymbols('\\varphi')
    dphi=dynamicsymbols('\\varphi',1)
    D=Symbol('D', positive=True)
    mu=Symbol('\\mu', positive=True)
    M_s=Symbol('M_s', positive=True)
    M_T=Symbol('M_T', positive=True)
    G=Symbol('G', positive=True)
    g=Symbol('g',positive=True)
    alpha=Symbol('\\alpha')
    pos1=phi
    qs=[phi]
    phi_1=phi
    phi_2=phi_1*i_1
    phi_3=phi_2*i_2
    
    def __init__(self,
                 I_s=None,
                 I_k=None,
                 I_1=None,
                 i_1=None,
                 I_2=None,
                 i_2=None,
                 I_3=None,
                 I_4=None,
                 I_b=None,
                 A=None,
                 B=None,
                 M_r=None,
                 I_r=None,
                 M_b=None,
                 ivar=Symbol('t'),
                 phi=None,
                 dphi=None,
                 D=None,
                 mu=None,
                 M_s=None,
                 M_T=None,
                 G=None,
                 g=None,
                 alpha=None,
                 phi_1=None,
                 phi_2=None,
                 phi_3=None,
                 qs=None,
                 **kwargs):
        

        
        if I_s is not None: self.I_s = I_s
        if I_k is not None: self.I_k = I_k
        if I_1 is not None: self.I_1 = I_1
        if i_1 is not None: self.i_1 = i_1
        if I_2 is not None: self.I_2 = I_2
        if I_3 is not None: self.I_3 = I_3
        if i_2 is not None: self.i_2 = i_2
        if I_4 is not None: self.I_4 = I_4
        if I_b is not None: self.I_b = I_b
        if D is not None: self.D = D
        if G is not None: self.G = G
        if g is not None: self.g = g
        if mu is not None: self.mu = mu
        if M_T is not None: self.M_T = M_T
        if M_s is not None: self.M_s = M_s
        if phi_1 is not None: self.phi_1 = phi_1
        if phi_2 is not None: self.phi_2 = phi_2
        if phi_3 is not None: self.phi_3 = phi_3
        if alpha is not None: self.alpha = alpha
        if A is not None: self.A=A
        if B is not None: self.B=B
        if M_r is not None: self.M_r = M_r
        if M_b is not None: self.M_b = M_b
        if I_r is not None: self.I_r = I_r
        if dphi is not None: self.dphi = dphi
        if phi_1 is not None: self.phi_1 = phi_1
        if phi_2 is not None: self.phi_2 = phi_2
        if phi_3 is not None: self.phi_3 = phi_3
        
        self._init_from_components(**kwargs)

    @property
    def components(self):

        components = {}

        self._engine = Disk(self.I_s, self.phi_1, qs=self.qs)
#         self.disk_k = Disk(I_k, phi_1, qs=qs)
        self._disk_1 = Disk(self.I_1, self.phi_1, qs=self.qs)
        self._disk_2 = Disk(self.I_2, self.phi_2, qs=self.qs)
        self._disk_3 = Disk(self.I_3, self.phi_2, qs=self.qs)
        self._disk_4 = Disk(self.I_4, self.phi_3, qs=self.qs)
        self._disk_B = Disk(self.I_b, self.phi_3, qs=self.qs)
        self._load = MaterialPoint(self.G/self.g, pos1=self.phi_3* self.D/2, qs=self.qs)
        self._friction_comp = GravitationalForce(self.G/self.g, self.g, pos1=self.phi_3* self.D/2 * cos(self.alpha)*self.mu , qs=self.qs)
        self._force = Force(-self.M_T,pos1=self.phi_3,qs=self.qs)
        self._drive = Force(self.M_s,pos1=self.phi_1,qs=self.qs)
        self._gravity_comp = GravitationalForce(self.G/self.g, self.g, pos1=self.phi_3* self.D/2 * sin(self.alpha) , qs=self.qs)


        components['engine'] = self._engine
        components['disk_1'] = self._disk_1
        components['disk_2'] = self._disk_2
        components['disk_3'] = self._disk_3
        components['disk_4'] = self._disk_4
        components['disk_B'] = self._disk_B
        components['load'] = self._load
        components['friction_comp'] = self._friction_comp
        components['force'] = self._force
        components['drive'] = self._drive
        components['gravity_comp'] = self._gravity_comp
        
        return components

    def symbols_description(self):
        self.sym_desc_dict = {
            self.I: r'crank moment of inertia',
            # self.k_beam: r'Beam stiffness',
            # self.g: r'gravitational field acceleration'
        }

        return self.sym_desc_dict

    def get_default_data(self):
        E, I_s,I_k,I_b,I_1,I_2,I_3,I_4,i_1, i_2,D,G,g,mu,M_T,M_s= symbols('E I_s I_k I_b I_1 I_2 I_3 I_4 i_1 i_2 D G g \mu M_T M_s', positive=True)
        
        default_data_dict = {
            self.I_s: [20,30,40],
            self.I_k: [20,30,40],
            self.I_1: [20,30,40],
            self.I_2: [20,30,40],
            self.I_3: [20,30,40],
            self.I_4: [20,30,40],
            self.I_b: [20,30,40],
            self.i_1: [S.One*3/2,S.One*2],
            self.i_2: [S.One*3/2,S.One*2],
            self.D: [10,20,30],
            self.G: [10,20,30],
            self.g: [10,20,30],
            self.mu : [0.1],
            self.M_T : [10,20,30],
            self.M_s : [10,20,30],
#             self.phi_1: [10,20,30],
#             self.phi_2: [10,20,30],
#             self.phi_3  : [10,20,30],
            self.alpha  : [10,20,30],
        }

        return default_data_dict
        
    def steady_angular_velocity(self):
        obj=self
        eoms_num = N(obj._eoms[0].subs(obj.M_s,obj.A-obj.B*obj.dphi).subs(obj._given_data),3)
        eom_sol = dsolve(eoms_num, obj.q[0], ics={obj.q[0].subs(obj.ivar, 0): 0, obj.q[0].diff(obj.ivar).subs(obj.ivar, 0): 0})
        omega_steady_state =  N(eom_sol.rhs.diff(obj.ivar).subs(obj.ivar,oo),3)
        return omega_steady_state
    
    def reduced_torque(self):
        obj=self
        
        red_tor = -(obj._eoms[0].doit().subs(obj.M_s,obj.A-obj.B*obj.dphi).subs([(obj.q[0].diff(obj.ivar,obj.ivar),0),(obj.q[0].diff(obj.ivar),0)]))
        return red_tor
    
    def delta_1(self):
        
        obj=self
        M_Z = obj._eoms[0].doit().subs(obj.M_s,obj.A-obj.B*obj.dphi).subs(obj._given_data).subs([(obj.q[0].diff(obj.ivar,obj.ivar),0),(obj.q[0].diff(obj.ivar),0)])
        delta_1 = 0.4*(obj.A-Abs(M_Z))/(self.steady_angular_velocity()**2*self.reduced_inertia())

        return delta_1

    def reduced_inertia(self):
        obj = self
        
        ans = obj.inertia_matrix()[0]

        return ans

    def flywheel_moment_of_inertia(self):
        obj=self
        delta_1=self.delta_1()
        I_r=self.reduced_inertia()

        ik=((((delta_1/0.02)-1)*I_r).n(5))
        return ik
    def startup_mass_velocity(self):
        obj=self
        st_val=obj.D/2*obj.phi_3
        return st_val

# metody do zatwierdzenia - karolina

    def reduced_system(self):
        obj=self
        new_sys=obj.subs(obj.M_s,obj.A-obj.B*obj.dphi)
        Ir=self.I_r
        Mr=self.M_r
        Mb=self.M_b
        ir=Eq(Ir,obj.reduced_inertia())
        mr=Eq(Mr,obj.reduced_torque()-obj.B*obj.dphi)
        mb=Eq(Mb,mr.rhs-(obj.A-obj.B*obj.dphi))
        reduced_ode=new_sys.subs(obj.I_s,solve(ir,obj.I_s)[0]).subs(obj.M_T,solve(mr,obj.M_T)[0])._ode_system.expand()
        return reduced_ode

    def startup_time(self):
        obj=self
        I_r=obj.reduced_inertia()
        ts_val=3*I_r/obj.B
        return ts_val

    def passive_torque(self):
        obj=self
        tp_val=(obj.M_T+obj.G*obj.D/2*(sin(obj.alpha)+obj.mu*cos(obj.alpha)))*obj.i_1*obj.i_2
        return tp_val

    def units(self):
        units_dict={
            self.I_s:ureg.kilogram*ureg.meter**2,
            self.I_k:ureg.kilogram*ureg.meter**2,
            self.I_1:ureg.kilogram*ureg.meter**2,
            self.I_2:ureg.kilogram*ureg.meter**2,
            self.I_3:ureg.kilogram*ureg.meter**2,
            self.I_4:ureg.kilogram*ureg.meter**2,
            self.I_b:ureg.kilogram*ureg.meter**2,
            self.q[0]:ureg.radian,
            self.D:ureg.meter,
            self.ivar:ureg.second,
            self.M_s:ureg.newton*ureg.meter,
            self.M_T:ureg.newton*ureg.meter,
            self.G:ureg.newton,
            self.g:ureg.meter/ureg.second**2,
            self.alpha:ureg.radian,
            self.A:ureg.newton*ureg.meter,
            self.B:(ureg.newton*ureg.meter*ureg.second)/ureg.radian,
           }
        return units_dict
    
class SDOFWinchSystemTest(ComposedSystem):


    scheme_name = 'winch_complete.png'
    real_name = 'winch_cross_sec.png'
    
    I_s=Symbol('I_s', positive=True)
    I_k=Symbol('I_k', positive=True)
    I_1=Symbol('I_1', positive=True)
    i_1=Symbol('i_1', positive=True)
    I_2=Symbol('I_2', positive=True)
    I_b=Symbol('I_b', positive=True)
    A=Symbol('A', positive=True)
    B=Symbol('B', positive=True)
    ivar=Symbol('t')
    phi=dynamicsymbols('\\varphi')
    dphi=dynamicsymbols('\\varphi',1)
    D=Symbol('D', positive=True)
    mu=Symbol('\\mu', positive=True)
    M_s=Symbol('M_s', positive=True)
    M_T=Symbol('M_T', positive=True)
    G=Symbol('G', positive=True)
    g=Symbol('g',positive=True)
    alpha=Symbol('\\alpha')
    pos1=phi
    qs=[phi]
    phi_1=phi
    phi_2=phi_1*i_1

    def __init__(self,
                 I_s=None,
                 I_k=None,
                 I_1=None,
                 i_1=None,
                 I_2=None,
                 I_b=None,
                 A=None,
                 B=None,
                 ivar=Symbol('t'),
                 phi=None,
                 dphi=None,
                 D=None,
                 mu=None,
                 M_s=None,
                 M_T=None,
                 G=None,
                 g=None,
                 alpha=None,
                 phi_1=None,
                 phi_2=None,
                 qs=None,
                 **kwargs):
        

        
        if I_s is not None: self.I_s = I_s
        if I_k is not None: self.I_k = I_k
        if I_1 is not None: self.I_1 = I_1
        if i_1 is not None: self.i_1 = i_1
        if I_2 is not None: self.I_2 = I_2
        if I_b is not None: self.I_b = I_b
        if D is not None: self.D = D
        if G is not None: self.G = G
        if g is not None: self.g = g
        if mu is not None: self.mu = mu
        if M_T is not None: self.M_T = M_T
        if M_s is not None: self.M_s = M_s
        if phi_1 is not None: self.phi_1 = phi_1
        if phi_2 is not None: self.phi_2 = phi_2
        if alpha is not None: self.alpha = alpha
        if A is not None: self.A=A
        if B is not None: self.B=B
        if dphi is not None: self.dphi = dphi
        if phi_1 is not None: self.phi_1 = phi_1
        if phi_2 is not None: self.phi_2 = phi_2
        self._init_from_components(**kwargs)

    @property
    def components(self):

        components = {}

        self._engine = Disk(self.I_s, self.phi_1, qs=self.qs)
#         self.disk_k = Disk(I_k, phi_1, qs=qs)
        self._disk_1 = Disk(self.I_1, self.phi_1, qs=self.qs)
        self._disk_2 = Disk(self.I_2, self.phi_2, qs=self.qs)
#         self._disk_3 = Disk(self.I_3, self.phi_2, qs=self.qs)
#         self._disk_4 = Disk(self.I_4, self.phi_3, qs=self.qs)
        self._disk_B = Disk(self.I_b, self.phi_2, qs=self.qs)
        self._load = MaterialPoint(self.G/self.g, pos1=self.phi_2* self.D/2, qs=self.qs)
#         self._friction_comp = GravitationalForce(self.G/self.g, self.g, pos1=self.phi_3* self.D/2 * cos(self.alpha)*self.mu , qs=self.qs)
        self._force = Force(-self.M_T,pos1=self.phi_2,qs=self.qs)
        self._drive = Force(self.M_s,pos1=self.phi_1,qs=self.qs)
        self._gravity_comp = GravitationalForce(self.G/self.g, self.g, pos1=self.phi_2* self.D/2 * sin(self.alpha) , qs=self.qs)


        components['engine'] = self._engine
        components['disk_1'] = self._disk_1
        components['disk_2'] = self._disk_2
#         components['disk_3'] = self._disk_3
#         components['disk_4'] = self._disk_4
        components['disk_B'] = self._disk_B
        components['load'] = self._load
#         components['friction_comp'] = self._friction_comp
        components['force'] = self._force
        components['drive'] = self._drive
        components['gravity_comp'] = self._gravity_comp
        
        return components

    def symbols_description(self):
        self.sym_desc_dict = {
            self.I: r'crank moment of inertia',
            # self.k_beam: r'Beam stiffness',
            # self.g: r'gravitational field acceleration'
        }

        return self.sym_desc_dict

    def get_default_data(self):
        E, I_s,I_k,I_b,I_1,I_2,I_3,I_4,i_1, i_2,D,G,g,mu,M_T,M_s= symbols('E I_s I_k I_b I_1 I_2 I_3 I_4 i_1 i_2 D G g \mu M_T M_s', positive=True)
        
        default_data_dict = {
            self.I_s: [20,30,40],
            self.I_k: [20,30,40],
            self.I_1: [20,30,40],
            self.I_2: [20,30,40],
            self.I_b: [20,30,40],
            self.i_1: [20,30,40],
            self.D: [10,20,30],
            self.G: [10,20,30],
            self.g: [10,20,30],
            self.mu : [10,20,30],
            self.M_T : [10,20,30],
            self.M_s : [10,20,30],
            self.phi_1: [10,20,30],
            self.phi_2: [10,20,30],
            self.alpha  : [10,20,30],
        }

        return default_data_dict
        
    def steady_angular_velocity(self):
        obj=self
        eoms_num = N(obj._eoms[0].subs(obj.M_s,obj.A-obj.B*obj.dphi).subs(obj._given_data),3)
        eom_sol = dsolve(eoms_num, obj.q[0], ics={obj.q[0].subs(obj.ivar, 0): 0, obj.q[0].diff(obj.ivar).subs(obj.ivar, 0): 0})
        omega_steady_state =  N(eom_sol.rhs.diff(obj.ivar).subs(obj.ivar,oo),3)
        return omega_steady_state
    
    def reduced_torque(self):
        obj=self
        
        red_tor = -(obj._eoms[0].doit().subs(obj.M_s,obj.A-obj.B*obj.dphi).subs([(obj.q[0].diff(obj.ivar,obj.ivar),0),(obj.q[0].diff(obj.ivar),0)]))
        return red_tor
    
    def delta_1(self):
        
        obj=self
        M_Z = obj._eoms[0].doit().subs(obj.M_s,obj.A-obj.B*obj.dphi).subs(obj._given_data).subs([(obj.q[0].diff(obj.ivar,obj.ivar),0),(obj.q[0].diff(obj.ivar),0)])
        delta_1 = 0.4*(obj.A-Abs(M_Z))/(self.steady_angular_velocity()**2*self.reduced_inertia())

        return delta_1

    def reduced_inertia(self):
        obj = self
        
        ans = obj.inertia_matrix()[0]

        return ans

    def flywheel_moment_of_inertia(self):
        obj=self
        delta_1=self.delta_1()
        I_r=self.reduced_inertia()

        ik=((((delta_1/0.02)-1)*I_r).n(5))
        return ik
    def startup_mass_velocity(self):
        obj=self
        st_val=obj.D/2*obj.phi_2
        return st_val

#DONE - Kuba
class SDOFDrivetrainVehicleSystem(ComposedSystem):


    #scheme_name = 'vehicle_drivetrain.png'
    #real_name = 'vehicle_drivetrain_real.PNG'

    R=Symbol('R')
    M_m=Symbol('M_m', positive=True)
    omega_m=Symbol('\\omega_m', positive=True)
    I_m=Symbol('I_m', positive=True)
    I=Symbol('I', positive=True)
    I_k=Symbol('I_k', positive=True)
    omega_1=Symbol('\\omega_1', positive=True)
    I_a=Symbol('I_a', positive=True)
    z_1=Symbol('z_1', positive=True)
    I_1=Symbol('I_1', positive=True)
    z_2=Symbol('z_2', positive=True)
    I_2=Symbol('I_2', positive=True)      
    I_b=Symbol('I_b', positive=True)
    z_3=Symbol('z_3', positive=True)
    I_3=Symbol('I_3', positive=True)
    z_4=Symbol('z_4', positive=True)
    I_4=Symbol('I_4', positive=True)
    omega_3=Symbol('omega_3', positive=True)
    I_c=Symbol('I_c',positive=True)
    I_d=Symbol('I_d', positive=True)
    I_w=Symbol('I_w', positive=True)
    r=Symbol('r', positive=True)
    m=Symbol('m', positive=True)
    i_1=Symbol('i_1',positive=True)
    i_2=Symbol('i_2',positive=True)
    A=Symbol('A', positive=True)
    B=Symbol('B', positive=True)
    phi=dynamicsymbols('\\varphi')
    phi_1=dynamicsymbols('\\varphi_1')
    phi_2=dynamicsymbols('\\varphi_2')
    phi_3=dynamicsymbols('\\varphi_3')
    dphi=dynamicsymbols('\\varphi',1)
    alpha=Symbol('alpha')
    c=Symbol('c')
    g=Symbol('g')    
    
    def __init__(self,
                 R=None,
                 M_m=None,
                 omega_m=None,
                 I_m=None,
                 I=None,
                 I_k=None,
                 omega_1=None,
                 I_a=None,
                 z_1=None,
                 I_1=None,
                 z_2=None,
                 I_2=None,
                 I_b=None,
                 z_3=None,
                 I_3=None,
                 z_4=None,
                 I_4=None,
                 omega_3=None,
                 I_c=None,
                 I_d=None,
                 I_w=None,
                 r=None,
                 m=None,
                 i_1=None,
                 i_2=None,
                 A=None,
                 B=None,
                 phi=None,
                 phi_1=None,
                 phi_2=None,
                 phi_3=None,
                 dphi=None,
                 alpha=None,
                 c=None,
                 g=None,
                 ivar=Symbol('t'),
                 **kwargs):
        


        if M_m is not None: self.M_m = M_m
        if omega_m is not None: self.omega_m = omega_m
        if I_1 is not None: self.I_1 = I_1
        if i_1 is not None: self.i_1 = i_1
        if I_2 is not None: self.I_2 = I_2
        if I_3 is not None: self.I_3 = I_3
        if i_2 is not None: self.i_2 = i_2
        if I_4 is not None: self.I_4 = I_4
        if I_a is not None: self.I_a = I_a
        if I_b is not None: self.I_b = I_b
        if I_c is not None: self.I_c = I_c
        if I_d is not None: self.I_d = I_d
        if I_k is not None: self.I_k = I_k
        if I_m is not None: self.I_m = I_m
        if I_w is not None: self.I_w = I_w
        if phi_1 is not None: self.phi_1 = phi_1
        if phi_2 is not None: self.phi_2 = phi_2
        if phi_3 is not None: self.phi_3 = phi_3
        if alpha is not None: self.alpha = alpha
        if A is not None: self.A = A
        if B is not None: self.B = B
        if dphi is not None: self.dphi = dphi
        if r is not None: self.r = r
#        if D is not None: self.D = 2*self.r
        if m is not None: self.m = m
        if c is not None: self.c = c
        if g is not None: self.g = g
        self.ivar=ivar
        self.qs = [self.phi] 

        self._init_from_components(**kwargs)
    #@cached_property  //not defined
    @property
    def components(self):
        components = {}
        
#        pos1=phi
        #qs=[self.phi]
        self.phi_1=self.phi
##         i_1=z_2/z_1
##         i_2=z_4/z_3
        self.phi_2=self.phi_1/self.i_1
        self.phi_3=self.phi_2/self.i_2
  
        
        self.engine_force = Force(self.M_m, pos1=self.phi_1,qs=[self.qs])
        self.engine_inertia = Disk(self.I_m, self.phi_1, qs=[self.qs])
#         self.flywheel = Disk(I, phi_1, qs=qs)
        self.clutch = Disk(self.I_c, self.phi_1, qs=[self.qs])
        self.driveshaft_1 = Disk(self.I_a, self.phi_1, qs=[self.qs])
        self.transmission_i = Disk(self.I_1, self.phi_1, qs=[self.qs])
        self.transmission_o = Disk(self.I_2, self.phi_2, qs=[self.qs])
        self.driveshaft_2 = Disk(self.I_b, self.phi_2, qs=[self.qs])
        self.diff_i = Disk(self.I_3, self.phi_2, qs=[self.qs])
        self.diff_o = Disk(self.I_4, self.phi_3, qs=[self.qs])
        self.axleshaft = Disk(self.I_d, self.phi_3, qs=[self.qs])
        self.wheels = Disk(4*self.I_w, self.phi_3, qs=[self.qs])
        self.mass = MaterialPoint(self.m, pos1=self.phi_3*self.r , qs=[self.qs])
        self.gravity = Force(-self.m*self.g*sin(self.alpha), pos1=self.phi_3*self.r,qs=[self.qs])#GravitationalForce(m, g, pos1=phi_3* r * sin(alpha) , qs=qs)
        self.air_force = Damper(self.c,pos1=self.phi_3*self.r,qs=[self.phi_1]) #(self, c, pos1, pos2=0, qs=None, ivar=Symbol('t'), frame=base_frame)

        components['_engine_force'] = self.engine_force
        components['_engine_inertia'] = self.engine_inertia
        components['_clutch'] = self.clutch
        components['_driveshaft_1'] = self.driveshaft_1
        components['_transmission_i'] = self.transmission_i
        components['_transmission_o'] = self.transmission_o
        components['_driveshaft_2'] = self.driveshaft_2
        components['_diff_i'] = self.diff_i
        components['_diff_o'] = self.diff_o
        components['_axleshaft'] = self.axleshaft
        components['_wheels'] = self.wheels
        components['_mass'] = self.mass
        components['_gravity'] = self.gravity
        components['_air_force'] = self.air_force


        
        return components                


    def get_default_data(self):
        
        E, I_s,I_k,I_b,I_1,I_2,I_3,I_4,i_1, i_2,D,G,g,mu,M_T,M_s= symbols('E I_s I_k I_b I_1 I_2 I_3 I_4 i_1 i_2 D G g \mu M_T M_s', positive=True)
        
        default_data_dict = {
#             self.I_s: [20,30,40],
#             self.I_k: [20,30,40],
#             self.I_1: [20,30,40],
#             self.I_2: [20,30,40],
#             self.I_3: [20,30,40],
#             self.I_4: [20,30,40],
#             self.I_b: [20,30,40],
#             self.i_1: [20,30,40],
#             self.i_2: [20,30,40],
#             self.D: [10,20,30],
#             self.G: [10,20,30],
#             self.g: [10,20,30],
#             self.mu : [10,20,30],
#             self.M_T : [10,20,30],
#             self.M_s : [10,20,30],
#             self.phi_1: [10,20,30],
#             self.phi_2: [10,20,30],
#             self.phi_3  : [10,20,30],
        }

        return default_data_dict
    
    def linear_vehicle_velocity(self):
        obj=self
        lin_veh_vel=omega_steady_state*vel_ratio
        return lin_veh_vel
    
    def velocity_ratio(self):
        obj=self
        vel_ratio=self.phi_3*self.r/self.phi
        return vel_ratio
    

    def steady_angular_velocity(self):
        obj=self


        eom_sol = obj._eom_solution()
        #omega_steady_state =  (eom_sol.rhs.diff(obj.ivar).subs(obj.ivar,oo))

        #display(obj._eoms[0])
        #eoms_num = N(obj._eoms[0].subs(obj.M_s,obj.A-obj.B*obj.dphi).subs(obj._given_data),3)
        #display(eoms_num)
        #eom_sol = dsolve(eoms_num, obj.q[0], ics={obj.q[0].subs(obj.ivar, 0): 0, obj.q[0].diff(obj.ivar).subs(obj.ivar, 0): 0})
        omega_steady_state =  N(eom_sol.rhs.diff(obj.ivar).subs(obj.ivar,oo),3)

        return omega_steady_state
    
    def steady_angular_velocity_rev_per_min(self):
        obj=self
        ang_vel = obj.steady_angular_velocity()
        return ang_vel* 60/360
    
    def reduced_torque(self):
        obj=self
        red_tor = (obj._eoms[0].doit().subs(obj.M_s,obj.A-obj.B*obj.dphi).subs([(obj.q[0].diff(obj.ivar,obj.ivar),0),(obj.q[0].diff(obj.ivar),0)]))
        return red_tor
    
    def delta_1(self):
        obj=self
        M_Z = obj._eoms[0].doit().subs(obj.M_s,obj.A-obj.B*obj.dphi).subs(obj._given_data).subs([(obj.q[0].diff(obj.ivar,obj.ivar),0),(obj.q[0].diff(obj.ivar),0)])
        delta_1 = 0.4*(obj.A-Abs(M_Z))/(self.steady_angular_velocity()**2*self.reduced_inertia())
#         0.4*(num_data.loc[case_no,'A']-Abs(M_Z))/(omega_ust**2*I_r)
        return delta_1

    def reduced_inertia(self):
        obj = self
        ans = obj.inertia_matrix()[0]
        return ans

    def flywheel_moment_of_inertia(self):
        obj=self
        delta_1=self.delta_1()
        I_r=self.reduced_inertia()

        ik=((((delta_1/0.004)-1)*I_r).n(5))
        return ik
    
    def startup_mass_velocity(self):
        obj=self
        st_val=obj.D/2*obj.phi_3
        return st_val
    
    def reducted_mass(self):
        obj=self
        
        return obj.flywheel_moment_of_inertia()/(obj.startup_mass_velocity()*obj.phi1)**2
    
    def reducted_force(self):
        obj=self
        
        return obj.reduced_torque()/obj.startup_mass_velocity()


    def _eom_solution(self):
        obj=self

        if obj.__cached_solution is None:
            eoms_num = N(obj._eoms[0].subs(obj.M_s,obj.A-obj.B*obj.dphi).subs(obj._given_data),3)
            EOM_dsolve = dsolve(eoms_num, obj.q[0], ics={obj.q[0].subs(obj.ivar, 0): 0, obj.q[0].diff(obj.ivar).subs(obj.ivar, 0): 0})
            obj.__cached_solution = EOM_dsolve
        else:
            EOM_dsolve = obj.__cached_solution

        return EOM_dsolve        
    



from sympy import (Symbol, symbols, Matrix, sin, cos, asin, diff, sqrt, S,
                   diag, Eq, hessian, Function, flatten, Tuple, im, pi, latex,
                   dsolve, solve, fraction, factorial, Subs, Number, oo, Abs,
                   N, solveset)

from sympy.physics.mechanics import dynamicsymbols, ReferenceFrame, Point
from sympy.physics.vector import vpprint, vlatex
from ...dynamics import LagrangesDynamicSystem, HarmonicOscillator, mech_comp

from ..elements import MaterialPoint, Spring, GravitationalForce, Disk, RigidBody2D, Damper, PID, Excitation, Force, base_frame, base_origin
from  ..continuous import ContinuousSystem, PlaneStressProblem

import base64
import random
import IPython as IP
import numpy as np
import inspect

from .principles import ComposedSystem, NonlinearComposedSystem, base_frame, base_origin, REPORT_COMPONENTS_LIST
from functools import cached_property


    
class UndampedVehicleSuspension(ComposedSystem):

    scheme_name = 'car_undamped.png'
    real_name = 'car_real.jpg'

    m=Symbol('m', positive=True)
    I=Symbol('I', positive=True)
    l_rod=Symbol('l_{rod}', positive=True)
    l_l=Symbol('l_l', positive=True)
    l_r=Symbol('l_r', positive=True)
    r=Symbol('k_r', positive=True)
    k_l=Symbol('k_l', positive=True)
    k_r=Symbol('k_r', positive=True)
    F_engine=Symbol('F_{engine}', positive=True)
    Omega=Symbol('Omega',positive=True)
    ivar=Symbol('t')
    #qs=dynamicsymbols('z, \\varphi')
    phi=dynamicsymbols('\\varphi')
    z=dynamicsymbols('z')
    #default symbols for substitution
    m0=Symbol('m_0', positive=True)
    l0=Symbol('l_0', positive=True)
    c0=Symbol('c_0', positive=True)
    k0=Symbol('k_0', positive=True)
    lam0=Symbol('\lambda_0', positive=True)
    Omega0=Symbol('Omega_0', positive=True)
    F0=Symbol('F_0', positive=True)

    def __init__(self,
                 m=None,
                 I=None,
                 l_rod=None,
                 l_r=None,
                 k_r=None,
                 k_l=None,
                 l_l=None,
                 F_engine=None,
                 Omega=None,
                 phi=None,
                 z=None,
                 ivar=Symbol('t'),
                 qs=None,
                 **kwargs):

        if  m is not None: self.m =m # mass of a rod
        if  I is not None: self.I = I # moment of inertia of a rod
        if  l_rod is not None: self.l_rod =l_rod# length of a rod
        if  l_r is not None: self.l_r = l_r 
        if  l_l is not None: self.l_l = l_l 
        if  k_r is not None: self.k_r =k_r# left spring
        if  k_l is not None: self.k_l = k_l# right spring
        if  F_engine is not None: self.F_engine = F_engine
        if  Omega is not None: self.Omega = Omega
        if  phi is not None: self.phi = phi
        if z is not None: self.z=z
        # self.z, self.phi = self.qs
        self.qs = [self.z,self.phi]

        self._init_from_components(**kwargs)

    @cached_property
    def components(self):

        components = {}

        self._body = RigidBody2D(self.m, self.I, pos_lin=self.z, pos_rot=self.phi, qs=self.qs)(label='rod')
        self._left_mount = Spring(self.k_l, pos1=self.z + self.phi * self.l_l, qs=self.qs)(label='left spring')
        self._right_mount = Spring(self.k_r, pos1=self.z - self.phi * self.l_r, qs=self.qs)(label='right spring')
        self._force = Force(self.F_engine*cos(self.Omega*self.ivar), pos1=self.z - self.l_r * self.phi, qs=self.qs)(label='force')
        self._static_force = Force(self.F_engine, pos1=self.z - self.l_r * self.phi, qs=self.qs)(label='force')

        components['_body'] = self._body
        components['_left_mount'] = self._left_mount
        components['_right_mount'] = self._right_mount
        components['_force'] = self._force
        components['_static_force'] = self._static_force

        return components

    def get_numerical_parameters(self):

        m, I, l_rod, l_r, l_l, k_r, k_l, F_engine, Omega, phi, z = self.m, self.I, self.l_rod, self.l_r, self.l_l, self.k_r, self.k_l, self.F_engine, self.Omega, self.phi, self.z
        
        get_numerical_parameters = {

            self.m: 0.5, 
            self.I: 20, 
            self.l_rod: 2,
            self.l_r: 9.81,
            self.l_l: 0.5, 
            self.k_r: 20, 
            self.k_l: 2,
            self.F_engine: 9.81,
            self.Omega: 2,
            #self.phi: 3,
            #self.z: 1,
        }
        return get_numerical_parameters

    def symbols_description(self):

        self.sym_desc_dict = {
            self.m: r'Mass of system',
            self.I: r'Moment of inertia',
            self.l_rod: r'Length of the rod',
            self.l_l: r'Length of the left side of the suspension',
            self.l_r: r'Length of the right side of the suspension',
            self.k_r: r'Right spring coefficient',
            self.k_l: r'Left spring coefficient ',
            self.F_engine: r'Force of engine',
            self.Omega: r'Excitation frequency ',
            self.phi: r'Angular position of suspension',
            self.z: r'Linear position of suspension',}
        return {**super().symbols_description(),**self.sym_desc_dict}

    def unit_dict(self):

        from pint import UnitRegistry
        ureg=UnitRegistry()

        unit_dict = {
            self.k_r: ureg.newton/ureg.meter,
            self.k_l: ureg.newton/ureg.meter,
            self.m: ureg.kilogram,
            self.I: ureg.kilogram*ureg.meter**2,
            self.F_engine: ureg.newton,
            self.l_l: ureg.meter,
            self.l_r: ureg.meter,
            self.l_rod: ureg.meter,
            self.Omega: ureg.radian / ureg.second,
            self.z: ureg.meter,
            self.z.diff(self.ivar): ureg.meter / ureg.second,
            self.z.diff(self.ivar, 2): ureg.meter / ureg.second**2,
            self.phi: ureg.radian,
            self.phi.diff(self.ivar): ureg.radian / ureg.second,
            self.phi.diff(self.ivar, 2): ureg.radian / ureg.second**2,
            self.ivar: ureg.second,
        }
        return unit_dict

    def max_static_force_pin(self):

        ans=self.static_force()
        return abs(ans)

    def static_force(self):
        data=self._given_data
        ans=self.dynamic_force()
        free_coeff=ans.subs({cos(self.Omega*self.ivar):0, sin(self.Omega*self.ivar):0}).subs(data)
        return (free_coeff)
    
    def static_force_pin_diameter(self):
        kt=Symbol('k_t', positive=True)
        Re=Symbol('R_e', positive=True)
        return ((4*self.max_static_force_pin())/(pi*kt*Re))**(1/2)

    def dynamic_force(self):
        
        amps = self._fodes_system.steady_solution.as_dict()
        force = self.components['_left_mount'].force().doit().expand()#.doit()
        data=self._given_data
        
        #display(abs(self.components['_spring_l'].force()))
        return (self.components['_left_mount'].force().subs(amps)).subs(data).expand().doit()
    
    def max_static_force(self):
        return self.static_force()

    def max_dynamic_force_pin(self):
        
        ans = self.dynamic_force()
        
        #display(abs(self.components['_spring_l'].force()))
        data=self._given_data
        sin_coeff=ans.coeff(sin(self.Omega*self.ivar)).subs(data)
        #display(sin_coeff)
        cos_coeff=ans.coeff(cos(self.Omega*self.ivar)).subs(data)
        #display(cos_coeff)
        free_coeff=ans.subs({cos(self.Omega*self.ivar):0, sin(self.Omega*self.ivar):0}).subs(data)
        #display(free_coeff)

        return sqrt(sin_coeff**2 + cos_coeff**2) + abs(free_coeff)
    
    def max_dynamic_force(self):
        return self.max_dynamic_force_pin()

    def dynamic_bearing_force(self):
        L=Symbol('L')#wymagana trwałość
        return self.max_dynamic_force_pin() * L**(S.One/3)
    
    def dynamic_force_pin_diameter(self):
        kt=Symbol('k_t', positive=True)
        Re=Symbol('R_e', positive=True)
        
        load = abs(self.max_dynamic_force_pin())
        
        return ((4*load)/(pi*kt*Re))**(1/2)
    
    def get_default_data(self):

        c0, k_0, l_l0, omega, F_0 = symbols('c_0 k_0 l_l0 Omega F_0', positive=True)
        m0, l0  = self.m0,self.l0
        c0, k0= self.c0,self.k0
        lam0=self.lam0
        Omega0=self.Omega0
        F0=self.F0
        Omega = self.Omega

        default_data_dict = {
            self.l_l: [self.l_r],
            self.l_rod:[l0*S.One*no for no in range(1, 8)],
            self.I: [S.One/12*self.m *self.l_rod**2],
            self.l_rod:[l0*S.One*no for no in range(1, 8)],
            #self.m: [m0,2*m0,3*m0,4*m0,5*m0,6*m0,7*m0,8*m0,9*m0],
            self.m: [m0*S.One*no for no in range(1,8)],
            #self.c_r: [2 * c0, 3 * c0, 4 * c0, 5 * c0, 6 * c0],
#             self.c_r: [self.lam*(self.k_r)],
            #self.k: [k0*S.One*no for no in range(1,8)],
#             self.lam: [lam0/10*S.One*no for no in range(1,8)],
            self.k_l: [k0*S.One*no for no in range(1,8)],
            self.k_r: [k0*S.One*no for no in range(1,8)],
            
            
#             self.c_l: [self.lam*(self.k_l)],
            self.l_r: [l0*S.One*no for no in range(1, 8)],
            #self.l_l: [l0*S.One*no for no in range(1, 8)],
           
#             self.Omega: [self.Omega],
            #self.F_engine: [4*no*F0*S.One*abs(cos(pi/2*no))  + F0*S.One*no*cos(self.Omega*self.ivar) for no in range(1,8)],
            self.F_engine: [4 * no * F0 * S.One for no in range(1,8)],
        }

        return default_data_dict
    
    @property
    def _report_components(self):
        
        comp_list=[
        *REPORT_COMPONENTS_LIST,
        mech_comp.MaxDynamicForce,
        mech_comp.MaxStaticForce,
        mech_comp.DynamicPinDiameter,
        mech_comp.StaticPinDiameter,
        mech_comp.SpringForce
        ]
        
        return comp_list
    
class UndampedSymmetricalVehicleSuspension(UndampedVehicleSuspension):

    scheme_name = 'car_undamped.png'
    real_name = 'car_real.jpg'

    m=Symbol('m', positive=True)
    I=Symbol('I', positive=True)
    l_rod=Symbol('l_{rod}', positive=True)
    l_l=Symbol('l_l', positive=True)
    l_r=Symbol('l_r', positive=True)
    r=Symbol('k_r', positive=True)
    k_l=Symbol('k_l', positive=True)
    k_r=Symbol('k_r', positive=True)
    F_engine=Symbol('F_{engine}', positive=True)
    Omega=Symbol('Omega',positive=True)
    ivar=Symbol('t')
    #qs=dynamicsymbols('z, \\varphi')
    phi=dynamicsymbols('\\varphi')
    z=dynamicsymbols('z')
    #default symbols for substitution
    m0=Symbol('m_0', positive=True)
    l0=Symbol('l_0', positive=True)
    c0=Symbol('c_0', positive=True)
    k0=Symbol('k_0', positive=True)
    lam0=Symbol('\lambda_0', positive=True)
    Omega0=Symbol('Omega_0', positive=True)
    F0=Symbol('F_0', positive=True)
    

    def __init__(self,
                 m=None,
                 I=None,
                 l_rod=None,
                 l_r=None,
                 k_r=None,
                 k_l=None,
                 l_l=None,
                 F_engine=None,
                 Omega=None,
                 phi=None,
                 z=None,
                 ivar=Symbol('t'),
                 qs=None,
                 **kwargs):

        if  m is not None: self.m =m # mass of a rod
        if  I is not None: self.I = I # moment of inertia of a rod
        if  l_rod is not None: self.l_rod =l_rod# length of a rod
        if  l_r is not None: self.l_r = l_r 
        if  l_l is not None: self.l_l = l_l 
        if  k_r is not None: self.k_r =k_r# left spring
        if  k_l is not None: self.k_l = k_l# right spring
        if  F_engine is not None: self.F_engine = F_engine
        if  Omega is not None: self.Omega = Omega
        if  phi is not None: self.phi = phi
        if z is not None: self.z=z
        # self.z, self.phi = self.qs
        self.qs = [self.z,self.phi]

        self._init_from_components(**kwargs)

    @cached_property
    def components(self):

        components = {}

        self._body = RigidBody2D(self.m, self.I, pos_lin=self.z, pos_rot=self.phi, qs=self.qs)(label='rod')
        self._left_mount = Spring(self.k_l, pos1=self.z + self.phi * self.l_l, qs=self.qs)(label='left spring')
        self._right_mount = Spring(self.k_r, pos1=self.z - self.phi * self.l_r, qs=self.qs)(label='right spring')
        self._force = Force(self.F_engine*cos(self.Omega*self.ivar), pos1=self.z - self.l_r * self.phi, qs=self.qs)(label='force')
        self._static_force = Force(self.F_engine, pos1=self.z - self.l_r * self.phi, qs=self.qs)(label='force')

        components['_body'] = self._body
        components['_left_mount'] = self._left_mount
        components['_right_monut'] = self._right_mount
        components['_force'] = self._force
        components['_static_force'] = self._static_force

        return components

    def get_default_data(self):

        #m0, l0, c0, k_0, l_l0, omega, F_0 = symbols('m_0 l_0 c_0 k_0 l_l0 Omega F_0', positive=True)
        c0, k_0, l_l0, omega, F_0 = symbols('c_0 k_0 l_l0 Omega F_0', positive=True)
        m0, l0  = self.m0,self.l0
        c0, k0= self.c0,self.k0
        lam0=self.lam0
        Omega0=self.Omega0
        F0=self.F0
        Omega = self.Omega

        default_data_dict = {
            self.l_l: [self.l_r],
            self.l_rod:[l0*S.One*no for no in range(1, 8)],
            self.I: [S.One/12*self.m *self.l_rod**2],
            self.m: [m0*S.One*no for no in range(1,50)],
            self.k_l: [self.k_r],
            self.k_r: [k0*S.One*no for no in range(1,40)],
            self.l_r: [l0*S.One*no for no in range(1, 8)],
#           self.Omega: [self.Omega],
            self.F_engine: [4 * no * F0 * S.One for no in range(1,8)],
        }

        return default_data_dict
    
    def reference_data_dict(self):

        default_data_dict = {
            self.l_l: [self.l_r],
#             self.I: [S.One/12*self.m *self.l_rod**2],
            self.k_l: [self.k_r],
        }

        return default_data_dict

    
class DampedVehicleSuspension(UndampedVehicleSuspension):

    scheme_name = 'damped_car.png'
    real_name = 'car_real.jpg'

    c_l=Symbol('c_l', positive=True)
    c_r=Symbol('c_r', positive=True)
                
    l_cl=Symbol('l_{cl}', positive=True)
    l_cr=Symbol('l_{cr}', positive=True)
    lam=Symbol('lambda',positive=True)
    phi=dynamicsymbols('\\varphi')
    z=dynamicsymbols('z')


    def __init__(self,
                 m=None,
                 I=None,
                 l_rod=None,
                 l_r=None,
                 k_r=None,
                 k_l=None,
                 F_engine=None,
                 c_l=None,
                 c_r=None,
                 l_cl=None,
                 l_l=None,
                 l_cr=None,
                 phi=None,
                 z=None,
                 ivar=Symbol('t'),
                 qs=None,
                 **kwargs):

        if  l_cr is not None: self.l_cr = l_cr
        if  l_cl is not None: self.l_cl = l_cl

        if  c_l is not None: self.c_l =c_l
        if  c_r is not None: self.c_r =  c_r


        self.qs = [self.phi,self.z]

        self._init_from_components(**kwargs)


    @cached_property
    def components(self):

        components =super().components

        
        self._damper_l = Damper(self.c_l, pos1=self.z + self.phi * self.l_l,
                               qs=self.qs)(label='left damper')
        self._damper_r = Damper(c=self.c_r, pos1=self.z - self.phi * self.l_r,
                               qs=self.qs)(label='right damper')



        components['_damper_l'] = self._damper_l
        components['_damper_r'] = self._damper_r
       

        
        return components

#     def max_static_force_pin(self):
#         data=self._given_data
#         ans=self.dynamic_force()
#         free_coeff=ans.subs({cos(self.Omega*self.ivar):0, sin(self.Omega*self.ivar):0}).subs(data)
#         return abs(free_coeff)
    
#     def static_force_pin_diameter(self):
#         kt=Symbol('k_t', positive=True)
#         Re=Symbol('R_e', positive=True)
#         return ((4*self.max_static_force_pin())/(pi*kt*Re))**(1/2)
    
#     def dynamic_force(self):
        
#         amps = self._fodes_system.steady_solution.as_dict()
#         force = self.components['_spring_l'].force().doit().expand()#.doit()
#         data=self._given_data
        
#         #display(abs(self.components['_spring_l'].force()))
#         return (self.components['_spring_l'].force().subs(amps)).subs(data).expand().doit()

#     def max_dynamic_force_pin(self):
        
#         ans = self.dynamic_force()
        
#         #display(abs(self.components['_spring_l'].force()))
#         data=self._given_data
#         sin_coeff=ans.coeff(sin(self.Omega*self.ivar)).subs(data)
#         #display(sincoeff)
#         cos_coeff=ans.coeff(cos(self.Omega*self.ivar)).subs(data)
#         #display(coscoeff)
#         free_coeff=ans.subs({cos(self.Omega*self.ivar):0, sin(self.Omega*self.ivar):0}).subs(data)
#         #display(freecoeff)

#         return sqrt(sin_coeff**2 + cos_coeff**2) + abs(free_coeff)

    
#    def dynamic_bearing_force(self):
       
#       return self.max_dynamic_force_pin() * self.sqrt(L)
    

#     def max_dynamic_force_pin(self):
#         return (self._frf()[0].subs(self.lam0,0) + self._frf()[1].subs(self.lam0,0)*self.l_rod/2)*self.stiffness_matrix()[0]

    def dynamic_bearing_force(self):
        L=Symbol('L')#wymagana trwałość
        return self.max_dynamic_force_pin() * L**(S.One/3)

    
    def dynamic_bearing_force(self):
        L=Symbol('L')
        return self.max_dynamic_force_pin() * self.sqrt(L)
    
    def symbols_description(self):
        self.sym_desc_dict = {
            self.m: r'Mass',
            self.I: r'',
            self.l_rod: r'',
            self.l_r: r'',
            self.k_r: r'',
            self.F_engine: r'',
            self.k_l: r'',
            self.c_l: r'',
            self.c_r: r'',
            self.l_cl: r'',
            self.l_l: r'',
            self.c_r: r'',
            self.phi: r'',
            self.Z: r'',
            self.ivar: r'',
            self.qs: r'',
        }
    
    def get_default_data(self):

        #m0, l0, c0, k_0, l_l0, omega, F_0 = symbols('m_0 l_0 c_0 k_0 l_l0 Omega F_0', positive=True)
        c0, k_0, l_l0, Omega0, F_0 = symbols('c_0 k_0 l_l0 Omega0 F_0', positive=True)
        m0, l0  = self.m0,self.l0
        c0, k0= self.c0,self.k0
        lam0=self.lam0
        Omega0=self.Omega0
        F0=self.F0

        default_data_dict = {
            self.l_rod:[l0*S.One*no for no in range(1, 8)],
            self.I: [S.One/12*self.m *self.l_rod],
            self.l_rod:[l0*S.One*no for no in range(1, 8)],
            #self.m: [m0,2*m0,3*m0,4*m0,5*m0,6*m0,7*m0,8*m0,9*m0],
            self.m: [m0*S.One*no for no in range(1,8)],
            #self.c_r: [2 * c0, 3 * c0, 4 * c0, 5 * c0, 6 * c0],
            self.c_r: [self.lam*(self.k_r)],
            #self.k: [k0*S.One*no for no in range(1,8)],
            #self.lam: [lam0/10*S.One*no for no in range(1,8)],
            self.k_l: [k0*S.One*no for no in range(1,8)],
            self.k_r: [k0*S.One*no for no in range(1,8)],
            
            
            self.c_l: [self.lam*(self.k_l)],
            self.l_r: [l0*S.One*no for no in range(1, 8)],
            self.l_l: [l0*S.One*no for no in range(1, 8)],
           
#             self.Omega: [Omega0*S.One*no for no in range(1,2)],
            self.F_engine: [F0*S.One*no for no in range(1,20)],}
            


        return default_data_dict
    
    @property
    def _report_components(self):
        
        comp_list=[
        *REPORT_COMPONENTS_LIST,
        mech_comp.DampedVibrationFrequency,
        mech_comp.MaxDynamicForce,
        mech_comp.DynamicForceComponent,
        mech_comp.MaxStaticForce,
        mech_comp.DynamicPinDiameter,
        mech_comp.DampingMatrixComponent
        ]
        
        return comp_list

class DampedSymmetricalVehicleSuspension(DampedVehicleSuspension):
    def get_default_data(self):

        #m0, l0, c0, k_0, l_l0, omega, F_0 = symbols('m_0 l_0 c_0 k_0 l_l0 Omega F_0', positive=True)
        c0, k_0, l_l0, Omega0, F_0 = symbols('c_0 k_0 l_l0 Omega0 F_0', positive=True)
        m0, l0  = self.m0,self.l0
        c0, k0= self.c0,self.k0
        lam0=self.lam0
        Omega0=self.Omega0
        F0=self.F0

        default_data_dict = {
            self.l_rod:[l0*S.One*no for no in range(1, 8)],
            self.I: [S.One/12*self.m *self.l_rod**2],
            self.l_rod:[l0*S.One*no for no in range(1, 8)],
            #self.m: [m0,2*m0,3*m0,4*m0,5*m0,6*m0,7*m0,8*m0,9*m0],
            self.m: [m0*S.One*no for no in range(1,8)],
            #self.c_r: [2 * c0, 3 * c0, 4 * c0, 5 * c0, 6 * c0],
            self.c_r: [self.lam*(self.k_r)],
            #self.k: [k0*S.One*no for no in range(1,8)],
            #self.lam: [lam0/10*S.One*no for no in range(1,8)],
            self.k_l: [self.k_r],
            self.k_r: [k0*S.One*no for no in range(1,8)],
            
            
            self.c_l: [self.lam*(self.k_l)],
            self.l_r: [self.l_l],
            self.l_l: [l0*S.One*no for no in range(1, 8)],
           
#             self.Omega: [Omega0*S.One*no for no in range(1,2)],
            self.F_engine: [F0 * S.One * no for no in range(1,20)],
        }

        return default_data_dict

    def get_random_parameters(self):

        default_data_dict = self.get_default_data()

        parameters_dict = {
            key: random.choice(items_list)
            for key, items_list in default_data_dict.items()
        }

        parameters_dict[self.k_l] = parameters_dict[self.k_r]
        parameters_dict[self.l_r] = parameters_dict[self.l_l]
#        parameters_dict[self.m_2] = parameters_dict[self.m]

        return parameters_dict

    def reference_data_dict(self):

        lam, m0, k0, l0, F0 = symbols('lambda m_0 k_0 l_0 F_0', positive=True)

        default_data_dict = {
            self.c_r: [self.lam*(self.k_r)],
            self.c_l: [self.lam*(self.k_l)],
            self.l_l: [self.l_r],
            self.k_l: [self.k_r],
        }

        return default_data_dict

class SimplifySuspension(UndampedVehicleSuspension):
    """
    Ready to use sample Double Degree of Freedom System represents symmetrical kinematically excited beam with two springs.
        Arguments:
        =========
            m = Mass
                -Mass of system on spring

            I = Moment of inertia
                -Moment of inertia of a rod

            l_rod = lenght of a rod
                -Dimension of a rod imitating vehicle's floor plate

            l_r = offset of right spring
                -Dimension of a right spring's offset

            l_l = offset of left spring
                -Dimension of a left spring's offset

            k_1 =Right spring coefficient
                -Right spring carrying the system

            k_2 =Left spring coefficient
                -Left spring carrying the system

            F_engine = thrust of an engine
                -Force made by engine, which allows vehicle to move forward

            ivar = symbol object
                -Independant time variable

            qs = dynamicsymbol object
                -Generalized coordinates

        Example
        =======
        A mass oscillating up and down while being held up by a spring with a spring constant kinematicly 

        >>> t = symbols('t')
        >>> m, k_1, k_2, l_rod, l_l, l_r, F_engine = symbols('m, k_1, k_2, l_rod, l_l, l_r, F_{engine}')
        >>> qs = dynamicsymbols('z, varphi') 
        >>> DDoFSymplifyVehicleSuspension()

        -We define the symbols and dynamicsymbols
        -and then we determine the instance of the system using class DDoFSymplifyVehicleSuspension()
    """
    scheme_name = 'vehicle_suspension.PNG'
    real_name = 'vehicle_suspension_real.PNG'

    m=Symbol('m', positive=True)
    I=Symbol('I', positive=True)
    l_rod=Symbol('l_{rod}', positive=True)
    l_l=Symbol('l_l', positive=True)
    l_r=Symbol('l_r', positive=True)
    k_1=Symbol('k_1', positive=True)
    k_2=Symbol('k_2', positive=True)
    F_engine=Symbol('F_{engine}', positive=True)
    Omega=Symbol('Omega', positive=True)
    ivar=Symbol('t')
    phi=dynamicsymbols('\\varphi')
    z=dynamicsymbols('z')

    def __init__(self,
                 m=None,
                 I=None,
                 l_rod=None,
                 l_r=None,
                 k_2=None,
                 k_1=None,
                 l_l=None,
                 F_engine=None,
                 phi=None,
                 z=None,
                 Omega=None,
                 ivar=Symbol('t'),
#                  qs=None,
                 **kwargs):

        if m is not None: self.m =m # mass of a rod
        if I is not None: self.I = I # moment of inertia of a rod
        if l_rod is not None: self.l_rod =l_rod# length of a rod
        if l_r is not None: self.l_r = l_r 
        if l_l is not None: self.l_l = l_l 
        if k_2 is not None: self.k_2 =k_2# left spring
        if k_1 is not None: self.k_1 = k_1# right spring
        if F_engine is not None: self.F_engine = F_engine
        if phi is not None: self.phi = phi
        if Omega is not None: self.Omega = Omega
        if z is not None: self.z=z
        self.ivar=ivar
        self.qs = [self.z,self.phi]

        self._init_from_components(**kwargs)


    @cached_property
    def components(self):

        components = {}

#         self._body = RigidBody2D(self.m, self.I, pos_lin=self.z, pos_rot=self.phi, qs=self.qs)(label='rod')
#         self._spring_l = Spring(self.k_1, pos1=self.z + self.phi * self.l_l, qs=self.qs)
#         self._spring_2 = Spring(self.k_2, pos1=self.z - self.phi * self.l_r, qs=self.qs)(label='right spring')
#         self._force = Force(self.F_engine*cos(self.Omega*self.ivar), pos1=self.z - self.phi * self.l_r, qs=self.qs)(label='force')

#         components['_body'] = self._body
#         components['_spring_l'] = self._spring_l
#         components['_spring_2'] = self._spring_2
#         components['_force'] = self._force

        self._vehicle = UndampedVehicleSuspension(m=self.m, I=self.I, l_rod=self.l_rod, l_r=self.l_r, k_r=self.k_2, k_l=self.k_1, l_l=self.l_l, F_engine=self.F_engine, Omega=self.Omega, phi=self.phi, z=self.z, ivar=Symbol('t'), qs=None)

        components['_vehicle'] = self._vehicle

        return components

    def get_default_data(self):

        m0, k0, l0 = symbols('m_0 k_0 l_0', positive=True)
        e,m_e = symbols('e m_e',positive=True)

        default_data_dict = {
            self.I: [(self.m / 12) * (2 * self.l_rod)**2],
            self.m: [m0 * S.One * no for no in range(10,21)],
            self.k_1: [k0 * S.One * no for no in range(100,201)],
            self.k_2: [k0 * S.One * no for no in range(100,201)],
            self.l_l: [l0 * S.One * no for no in range(1,11)],
            self.l_r: [l0 * S.One * no for no in range(1,11)],
            self.l_rod: [l0 * S.One * no for no in range(3,7)],
            self.F_engine: [e*m_e*self.Omega**2],
            m_e: [m0 * S.One * no for no in range(10,11)],
            e: [l0 * S.One * no for no in range(1,6)],
        }

        return default_data_dict
    
class QuarterOfVehicle(ComposedSystem):
    
    k_d=Symbol('k_d', positive=True)
    k_u=Symbol('k_u', positive=True)
    m_d=Symbol('m_d', positive=True)
    m_u=Symbol('m_u', positive=True)
    c=Symbol('c', positive=True)
    omega=Symbol('omega', positive=True)
    A_0=Symbol('A_0', positive=True)
    x_1=dynamicsymbols('x_1')
    x_2=dynamicsymbols('x_2')
    ivar=Symbol('t')
    
    def __init__(self,
                 k_d=None,
                 k_u=None,
                 m_d=None,
                 m_u=None,
                 c=None,
                 omega=None,
                 A_0=None,
                 x_1=None,
                 x_2=None,
                 ivar=Symbol('t'),
                 **kwargs):
    
    
        if m_d is not None: self.m_d = m_d
        if k_d is not None: self.k_d = k_d
        if m_u is not None: self.m_u = m_u
        if k_u is not None:self.k_u = k_u
        if c is not None: self.c=c
        if omega is not None: self.omega=omega
        if A_0 is not None: self.A_0=A_0
        if x_1 is not None: self.x_1 = x_1
        if x_2 is not None : self.x_2 = x_2

        self.ivar = ivar
        self.qs = [self.x_1, self.x_2]

        self._init_from_components(**kwargs)
        
    @cached_property
    def components(self):
        
        components = {}
        
        self._material_point_1 = MaterialPoint(self.m_d, self.x_1, qs=self.qs)
        self._spring_1 = Spring(self.k_d,self.x_1,self.A_0*sin(self.ivar*self.omega),qs=self.qs)
        self._material_point_2 = MaterialPoint(self.m_u, self.x_2, qs=self.qs)
        self._spring_2 = Spring(self.k_u,self.x_1,self.x_2,qs=self.qs)
        self._damper_1=Damper(self.c,self.x_1,self.x_2, qs=self.qs)

        components['material_point_1'] = self._material_point_1
        components['spring_1'] = self._spring_1
        components['material_point_2'] = self._material_point_2
        components['spring_2'] = self._spring_2
        components['damper'] = self._damper_1
        
        return components
       
    def symbols_description(self):
        self.sym_desc_dict = {
            self.m_d: r'unsprung mass',
            self.m_u: r'sprung mass',
            self.k_d: r'wheel spring stiffness coefficient',
            self.k_u: r'body spring stiffness coefficient',
            self.c: r'współczynnik tłumienia zawieszenia',
            self.A_0: r'amplitude',
            self.x_1: r'wheel displacement',
            self.x_2: r'body displacement',
            self.x_1.diff(t): r'wheel velocity',
            self.x_2.diff(t): r'body velocity',
            self.omega: r'circular frequency',
            self.ivar: r'time',
            
        }
        return self.sym_desc_dict 

